#!/usr/bin/env python

"""
Table of contents macro.

Copyright (C) 2018 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from moinformat.macros.common import Macro
from moinformat.tree.moin import Container, Heading, List, ListItem, Text

class TableOfContents(Macro):

    "A table of contents macro."

    name = "TableOfContents"

    def evaluate(self):

        "Evaluate the macro, producing a table of contents."

        arglist = []
        _defaults = [None] * 2

        for arg, default in map(None, self.node.args, _defaults):
            if arg is not None:
                try:
                    arg = max(1, int(arg.strip()))
                except ValueError:
                    arg = None
            arglist.append(arg)

        self.make_table(arglist[0], arglist[1])

    def make_table(self, min_level=None, max_level=None):

        """
        Make a table of contents with the given 'min_level' and 'max_level' of
        headings.
        """

        headings = []
        self.find_headings(self.doc, headings)

        if not headings:
            return

        # Common list features.

        marker = "1."
        space = " "
        num = "1"
        nl = [Text("\n")]

        # Start with no lists, no current item.

        lists = []
        item = None
        level = 0

        for heading in headings:
            new_level = heading.level

            # Create new lists if the level increases.

            if new_level > level:
                while level < new_level:
                    level += 1

                    if not (min_level <= level <= max_level):
                        continue

                    # Determine whether the heading should be generated at this
                    # level.

                    nodes = level == new_level and heading.nodes[:] + nl or []
                    indent = level - 1

                    # Make a list and add an item to it.

                    new_items = []
                    new_list = List(new_items, indent, marker, num)
                    new_item = ListItem(nodes, indent, marker, space, None)
                    new_items.append(new_item)

                    # Add the list to the current item, if any.

                    if item:
                        item.nodes.append(new_list)

                    # Record the new list.

                    lists.append(new_list)

                    # Reference the new list's items and current item.

                    items = new_items
                    item = new_item

            else:
                # Retrieve an existing list if the level decreases.

                if new_level < level:
                    while level > new_level:
                        if min_level <= level <= max_level:
                            lists.pop()
                        level -= 1

                    # Obtain the existing list and the current item.

                    items = lists[-1].nodes
                    item = items[-1]

                # Add the heading as an item.

                if min_level <= level <= max_level:
                    indent = level - 1
                    nodes = heading.nodes[:] + nl

                    item = ListItem(nodes, indent, marker, space, None)
                    items.append(item)

        # Replace the macro node's children with the top-level list.

        self.node.nodes = [lists[0]]

    def find_headings(self, node, headings):

        "Find headings under 'node', adding them to the 'headings' list."

        if node.nodes:
            for n in node.nodes:
                if isinstance(n, Heading):
                    headings.append(n)
                elif isinstance(n, Container):
                    self.find_headings(n, headings)

macro = TableOfContents

# vim: tabstop=4 expandtab shiftwidth=4
